/*
 * main.c: MPI ring-decomposition N-body code driver.
 *
 * Michael S. Warren, Sep. 11, 1995.  mswarren@lanl.gov
 */

#include <mpi.h>
#include <stdarg.h>
#include <stdlib.h>
#include "ring.h"
#include "grav.h"
#include "utils.h"
#include "rand.h"
#include "vop.h"

typedef struct {
    float mass;			/* mass of body */
    float pos[NDIM];		/* position of body */
    float vel[NDIM];		/* velocity of body */
    float acc[NDIM];
    float phi;
} body;

/* When we send a body from node to node, how much must we send??? */
#define TBODYSZ (1+NDIM)*sizeof(float)

void Diags(body *btab, int nobj, double tm);
void set_eps(float eps);
void set_body(void *o, void *p);
void ring_grav(void *p0, void *list, int bsize, int n);

void
main(int argc, char *argv[])
{
    int procnum, nproc;
    body *p, *btab;
    int gnobj;
    int nobj;
    double starttime, endtime; 

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &procnum);
    MPI_Comm_size (MPI_COMM_WORLD, &nproc);
    singlPrintf("Welcome to the MPI N-body code on %d procs\n", nproc);

    starttime = MPI_Wtime(); 

    /* Use command line arg to set N, if specified */
    if (argc == 2) gnobj = atoi(argv[1]);
    else gnobj = 1024;

    nobj = gnobj/nproc;
    btab = Malloc(nobj * sizeof(body));

    /* Generate random dataset. */
    ranset(17+procnum);
    for (p = btab; p < btab+nobj; p++) {
	p->mass = 1.0 / gnobj;
	sphere_rand(NDIM, p->pos);
	sphere_rand(NDIM, p->vel);
	VS(p->acc, = 0.0);
	p->phi = 0.0;
    }
    set_eps(.01);		/* set force smoothing length */
    singlPrintf("Calling ring...\n");

    /* Interact btab with itself */
    Ring(btab, sizeof(body), nobj, btab, sizeof(body), nobj, 
	 TBODYSZ, set_body, ring_grav);

    endtime = MPI_Wtime(); 

    Diags(btab, nobj, endtime-starttime);

    Free(btab);
    singlPrintf("Bye.\n");
    MPI_Finalize();
}

void
Diags(body *btab, int nobj, double tm) /* Energy diagnostics. */
{
    body *p;
    int gnobj;
    float pe, ke;
    double max_tm;
    float local_pe = 0.0;
    float local_ke = 0.0;

    for (p = btab; p < btab+nobj; p++) {
	local_pe += 0.5 * p->mass * p->phi;
	local_ke += 0.5 * p->mass * Dot(p->vel, p->vel);
    }
    
    MPI_Reduce(&tm, &max_tm, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
    MPI_Reduce(&local_pe, &pe, 1, MPI_FLOAT, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(&local_ke, &ke, 1, MPI_FLOAT, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(&nobj, &gnobj, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);

    singlPrintf("Total nobj is %d\n", gnobj);
    singlPrintf("Total potential energy is %12.8f\n", pe);
    singlPrintf("Total kinetic   energy is %12.8f\n", ke);
    singlPrintf("Wall clock time: %.2f seconds\n", max_tm);
}

static float Eps2;

void
set_eps(float eps)
{
    Eps2 = eps*eps;
}

void
set_body(void *o, void *p)	/* passed to ring for initialization */
{
    memcpy(o, p, TBODYSZ);
}

void 
ring_grav(void *p0, void *list, int bsize, int n)
{
    body *p = p0;
    float total_mass = 0.0;
    char *clist = (char *)list;

    do_grav(list, (float *)(clist + n * bsize), 
	    p->pos, &total_mass, p->acc, &p->phi, &Eps2);
}
