/* 
 * ring.c: Generic parallel ring topology code.  Use for any "all-pairs"
 * type of interaction.  Interacts two lists with each other, using
 * the function passed as interactf() to perform interactions.
 *
 * Michael S. Warren, Sep. 11, 1995.  mswarren@lanl.gov
 */

#include <mpi.h>
#include "utils.h"

#define MSGTAG 101

void
Ring(void *bptr, int bsize, int bnobj,
     void *optr, int osize, int onobj, int oused, 
     void initf(void *, void *), void interactf(void *, void *, int, int))
{
    char *p;
    void *travel_btab;
    void *tmpbuf;
    int travel_size;
    int n, i;
    int from_proc, to_proc;
    int procnum;
    int nproc;
    int max_nobj = onobj;

    MPI_Comm_rank(MPI_COMM_WORLD, &procnum);
    MPI_Comm_size (MPI_COMM_WORLD, &nproc);

    /* Determine size of largest list */
    MPI_Reduce(&onobj, &max_nobj, 1, MPI_INT, MPI_MAX, 0, MPI_COMM_WORLD);

    travel_size = max_nobj * oused;
    travel_btab = Malloc(travel_size);
    tmpbuf = Malloc(travel_size); /* communication buffer */

    /* Initialize the list that will be passed around the ring. */
    for (i = 0; i < onobj; i++) {
	p = (char *)optr+i*osize;
	initf((char *)travel_btab+i*oused, p);
    }

    /* define ring topology */
    to_proc = (procnum+1)%nproc;
    from_proc = (procnum+nproc-1)%nproc;

    /* local part */
    for (p = bptr; p < (char *)bptr + bnobj * bsize; p += bsize) {
	interactf(p, travel_btab, oused, onobj);
    }

    for(n = 1; n < nproc; n++) {
	MPI_Request req1, req2;
	MPI_Status stat;
	singlPrintf("cycle %d starting\n", n);

	memcpy(tmpbuf, travel_btab, travel_size);
	MPI_Irecv(travel_btab, travel_size, MPI_BYTE, from_proc, MSGTAG, 
		 MPI_COMM_WORLD, &req1);
	MPI_Isend(tmpbuf, onobj * oused, MPI_BYTE, to_proc, MSGTAG, 
		 MPI_COMM_WORLD, &req2);
	MPI_Wait(&req2, &stat);
	MPI_Wait(&req1, &stat);
	MPI_Get_count(&stat, MPI_BYTE, &onobj);
	onobj /= oused;

	for (p = bptr; p < (char *)bptr + bnobj * bsize; p += bsize) {
	    interactf(p, travel_btab, oused, onobj);
	}
    }
    Free(tmpbuf);
    Free(travel_btab);
}

